﻿// Copyright (c) Vitaly Kamiansky. All rights reserved. 
// See License.txt in the project root.

using Property;
using Property.Windows;
using Property.Windows.Legacy;
using System;
using System.Linq;
using System.Collections.Generic;
using System.Threading;
using System.Windows.Input;

namespace Showcase.LoadByWire
{
    // Demonstrates Load-By-Wire pure Rx-style version
    // Many subscriptions, many things to dispose. 
    // However, it looks scarier than it is.
    // ...
    // Still can't figure out? A hint: the actuator 
    // serves as a two-mode switch: in one mode user input goes to output
    // and there's no feedback, the other mode is exactly the opposite.
    public class MainWindowViewModel
    {
        public MainWindowViewModel()
        {
            // Declare shortcuts
            var throttlePattern = Reloadable<string>.On().Delayed(TimeSpan.FromMilliseconds(200));

            // Describe properties 
            var passportNr = throttlePattern.CheckInput(o => ValidationService.GetPassportNrError(o)).Create("");
            var firstName = throttlePattern.CheckInput(o => ValidationService.GetFirstNameError(o)).Create("");
            var lastName = throttlePattern.CheckInput(o => ValidationService.GetLastNameError(o)).Create("");
            var company = throttlePattern.CheckInput(o => ValidationService.GetCompanyError(o)).Create("");
            var email = throttlePattern.CheckInput(o => ValidationService.GetEmailError(o)).Create("");

            var user = Reloadable<Employee>.On().Each().Get(_ => new Employee(
                passportNr.Value == _Result.Value.PassportNr ? _Result.Value.Id : -1,
                passportNr.Value,
                firstName.Value,
                lastName.Value,
                company.Value,
                email.Value)).Create(new Employee(-1, "", "", "", "", ""));

            var dbMode = new Actuator();

            var db = Reloadable<Employee>.On().Worker().Each().Call(_ =>
                GetFakeEmployee(passportNr.Value)).Create();

            var status = Reloadable<LoadingStatus>.On().Each().Call(_ =>
                dbMode.Power.Value ? db.IsLoading ? LoadingStatus.InProcess :
                                                    db.Value == null ? LoadingStatus.EmptyResultSet :
                                                                       LoadingStatus.RecordsFound :
                                     LoadingStatus.NoActivity).Create();

            var result = Reloadable<Employee>.On().Each().Call(o =>
                dbMode.Power.Value ? db.Value ?? o : user.Value)
                .Create(new Employee(-1, "", "", "", "", ""));

            var dbToResult = db.OnChanged(() =>
                dbMode.Power.Value ? result : null);

            var resultToStatus = result.OnChanged(() =>
                status);

            var feedback = result.OnChanged(() =>
            {
                if (!dbMode.Power.Value) return;

                passportNr.Input = result.Value.PassportNr;
                firstName.Input = result.Value.FirstName;
                lastName.Input = result.Value.LastName;
                company.Input = result.Value.Company;
                email.Input = result.Value.Email;

                status.Go();
                dbMode.Stop();
            });


            var passportNrToResult = passportNr.OnChanged(() => result.Value.PassportNr != passportNr.Value ? result : null);
            var firstNameToResult = firstName.OnChanged(() => result.Value.FirstName != firstName.Value ? result : null);
            var lastNameToResult = lastName.OnChanged(() => result.Value.LastName != lastName.Value ? result : null);
            var companyToResult = company.OnChanged(() => result.Value.Company != company.Value ? result : null);
            var emailToResult = email.OnChanged(() => result.Value.Email != email.Value ? result : null);

            //=//

            // Describe what to show
            _PassportNr = passportNr;
            _FirstName = firstName;
            _LastName = lastName;
            _Company = company;
            _Email = email;
            _Status = status;
            _Result = result;

            _FindCommand = new DelegateCommand(_ =>
            {
                dbMode.Start();
                db.Go();
                status.Go();
            });

            // The disposal we keep in mind.
            _ToDispose = new[] { 
                (IDisposable)passportNr, firstName, lastName, company, email, 
                db, passportNrToResult, firstNameToResult, lastNameToResult,
                companyToResult,emailToResult };
        }

        //=//

        private readonly ICheckInputProperty<string> _PassportNr;
        public ICheckInputProperty<string> PassportNr { get { return _PassportNr; } }

        private readonly ICheckInputProperty<string> _FirstName;
        public ICheckInputProperty<string> FirstName { get { return _FirstName; } }

        private readonly ICheckInputProperty<string> _LastName;
        public ICheckInputProperty<string> LastName { get { return _LastName; } }

        private readonly ICheckInputProperty<string> _Company;
        public ICheckInputProperty<string> Company { get { return _Company; } }

        private readonly ICheckInputProperty<string> _Email;
        public ICheckInputProperty<string> Email { get { return _Email; } }

        private readonly ICommand _FindCommand;
        public ICommand FindCommand { get { return _FindCommand; } }

        private readonly IProperty<Employee> _Result;
        public IProperty<Employee> Result { get { return _Result; } }

        private readonly IProperty<LoadingStatus> _Status;
        public IProperty<LoadingStatus> Status { get { return _Status; } }


        // Instead of this in real world we would call an injected service
        private Employee GetFakeEmployee(string passportNr)
        {
            Thread.Sleep(1000);
            if (passportNr == "1111")
                return new Employee(2, "1111", "Me", "No Name", "Fake Company", "me@nomail.org");
            else
                return null;
        }

        //=//

        private readonly IList<IDisposable> _ToDispose;

        public void Dispose()
        {
            _ToDispose.ToList().ForEach(o => o.Dispose()); ;
        }
    }
}
