﻿// Copyright (c) Vitaly Kamiansky. All rights reserved. 
// See License.txt in the project root.

using Moq;
using System;
using System.Collections.Generic;
using System.Linq;

namespace Property.Test
{
    public abstract class PropertyPrototypeTestInfrastructure<T>
    {
        private Mock<IReloader> ReloaderMockState;
        protected Mock<IReloader> ReloaderMock
        {
            get
            {
                if (ReloaderMockState == null)
                {
                    ReloaderMockState = new Mock<IReloader>(MockBehavior.Strict);
                    ReloaderMockState.Setup(o => o.Go())
                        .Callback(() =>
                        {
                            DoReloaderBehaviour(GetValueState, ApplyValueState, ApplyStatusState);
                        });
                    ReloaderMockState.Setup(o => o.Dispose());
                }
                return ReloaderMockState;
            }
        }

        protected virtual bool TestStatus { get { return true; } }

        protected virtual void DoReloaderBehaviour(Func<T> getValue, Action<T> applyValue, Action<bool> applyStatus)
        {
            applyStatus(TestStatus);
            applyValue(getValue());
        }

        private Func<T> GetValueState;
        private Action<T> ApplyValueState;
        private Action<bool> ApplyStatusState;

        private Mock<IReloaderPrototype<T>> ReloaderPrototypeMockState;
        protected Mock<IReloaderPrototype<T>> ReloaderPrototypeMock
        {
            get
            {
                if (ReloaderPrototypeMockState == null)
                {
                    ReloaderPrototypeMockState = new Mock<IReloaderPrototype<T>>(MockBehavior.Strict);
                    ReloaderPrototypeMockState.Setup(o => o.Create(
                        It.IsAny<Func<T>>(),
                        It.IsAny<Action<T>>(),
                        It.IsAny<Action<bool>>()))
                        .Callback<Func<T>, Action<T>, Action<bool>>((o1, o2, o3) =>
                        {
                            GetValueState = o1;
                            ApplyValueState = o2;
                            ApplyStatusState = o3;
                        })
                        .Returns(ReloaderMock.Object);
                }
                return ReloaderPrototypeMockState;
            }
        }

        private Mock<IReloaderPrototype<T>> NullReloaderPrototypeMockState;
        protected Mock<IReloaderPrototype<T>> NullReloaderPrototypeMock
        {
            get
            {
                if (NullReloaderPrototypeMockState == null)
                {
                    NullReloaderPrototypeMockState = new Mock<IReloaderPrototype<T>>(MockBehavior.Strict);
                    NullReloaderPrototypeMockState.Setup(o => o.Create(
                        It.IsAny<Func<T>>(),
                        It.IsAny<Action<T>>(),
                        It.IsAny<Action<bool>>()))
                        .Returns<IReloader>(null);
                }
                return NullReloaderPrototypeMockState;
            }
        }

        protected virtual void SetupEvents(IProperty<T> sutProperty)
        {
            sutProperty.PropertyChanged += (o, e) =>
            {
                ResultEventStrings = ResultEventStrings.Concat(new[] { String.Format("Changed'{0}'", e.PropertyName) });
            };

            sutProperty.PropertyChanging += (o, e) =>
            {
                ResultEventStrings = ResultEventStrings.Concat(new[] { String.Format("Changing'{0}'", e.PropertyName) });
            };
        }

        private IEnumerable<string> ResultEventStrings 
            = Enumerable.Empty<string>();

        protected virtual IEnumerable<string> TestEventStrings
        { get { return new[] { 
            "Changing'IsLoading'", 
            "Changed'IsLoading'", 
            "Changing'Value'", 
            "Changed'Value'" }; } }


        //=//

        protected virtual void VerifyEvents()
        {
            new[] { ResultEventStrings.SequenceEqual(TestEventStrings) }
                .Where(o => !o)
                .ToList()
                .ForEach(o => { throw new Exception(); });
        }

        protected virtual void VerifyCreate()
        {
            ReloaderPrototypeMock.Verify(o => o.Create(
                It.IsAny<Func<T>>(),
                It.IsAny<Action<T>>(),
                It.IsAny<Action<bool>>()),
                Times.Once());
        }

        protected virtual void VerifyReload(int times)
        {
            VerifyCreate();
            ReloaderMock.Verify(o => o.Go(), Times.Exactly(times));
        }

        protected virtual void VerifyReload()
        {
            VerifyReload(1);
        }

        protected virtual void VerifyDispose()
        {
            VerifyCreate();
            ReloaderMock.Verify(o => o.Dispose(), Times.Once());
        }
    }
}
