﻿// Copyright (c) Vitaly Kamiansky. All rights reserved. 
// See License.txt in the project root.

using Property;
using Property.Windows;
using System;
using System.Linq;
using System.Collections.Generic;
using System.Reactive.Linq;
using System.Threading;
using System.Windows.Input;

namespace Showcase.LoadByWire
{
    // Demonstrates Load-By-Wire pure Rx-style version
    public class MainWindowViewModel
    {
        public MainWindowViewModel()
        {
            // Declare shortcuts
            var throttlePattern = Reloadable<string>.On().Throttle(TimeSpan.FromMilliseconds(200));
            var inputPattern = Reloadable<Employee>.On().Each().Input();

            // Describe properties 
            var passportNr = throttlePattern.ValidationInput(o => ValidationService.GetPassportNrErrors(o)).Create("");
            var firstName = throttlePattern.ValidationInput(o => ValidationService.GetFirstNameErrors(o)).Create("");
            var lastName = throttlePattern.ValidationInput(o => ValidationService.GetLastNameErrors(o)).Create("");
            var company = throttlePattern.ValidationInput(o => ValidationService.GetCompanyErrors(o)).Create("");
            var email = throttlePattern.ValidationInput(o => ValidationService.GetEmailErrors(o)).Create("");

            var user = inputPattern.Create(new Employee(-1, "", "", "", "", ""));
            var result = inputPattern.Create(new Employee(-1, "", "", "", "", ""));

            var db = Reloadable<Employee>.On().Async().Each().Call(_ => GetFakeEmployee(passportNr.Value)).Create();
            var status = Reloadable<LoadingStatus>.On().Each().Input().Create();
            var dbMode = new Actuator();

            // Describe subscriptions
            var controlsToUser = user.FromObservable(
                Observable.CombineLatest(
                Observable.Return("").Concat(passportNr.ToObservable()),
                Observable.Return("").Concat(firstName.ToObservable()),
                Observable.Return("").Concat(lastName.ToObservable()),
                Observable.Return("").Concat(company.ToObservable()),
                Observable.Return("").Concat(email.ToObservable()),
                (o1, o2, o3, o4, o5) => new Employee(
                    o1 == result.Value.PassportNr ? result.Value.Id : -1,
                    o1, o2, o3, o4, o5)));

            var userToMode = user.ToObservable().Where(o => o.Equals(result.Value))
                .Subscribe(_ => dbMode.Stop());

            var dbModeToStatus = status.FromObservable(dbMode.Power.ToObservable()
                .Select(o => o ? LoadingStatus.InProcess :
                    (db.Value == null ? LoadingStatus.EmptyResultSet :
                    LoadingStatus.RecordsFound)));

            var resultToStatus = status.FromObservable(result.ToObservable()
                .Where(_ => !dbMode.Power.Value).Select(_ => LoadingStatus.NoActivity));

            var dbModeToDb = db.FromObservable(dbMode.Power.ToObservable().Where(o => o));

            var dbToResult = result.FromObservable(db.ToObservable()
                .Where(_ => dbMode.Power.Value)
                .Select(o => o != null ? o : result.Value));

            var userToResult = result.FromObservable(user.ToObservable()
                .Where(o => !dbMode.Power.Value && !o.Equals(result.Value)));

            //=//

            var resultFeedback = result.ToObservable().Where(o => dbMode.Power.Value);
            var resultToPassportNr = passportNr.FromObservable(resultFeedback.Select(o => o.PassportNr));
            var resultToFirstName = firstName.FromObservable(resultFeedback.Select(o => o.FirstName));
            var resultToLastName = lastName.FromObservable(resultFeedback.Select(o => o.LastName));
            var resultToCompany = company.FromObservable(resultFeedback.Select(o => o.Company));
            var resultToEmail = email.FromObservable(resultFeedback.Select(o => o.Email));

            // Describe what to show

            _PassportNr = passportNr;
            _FirstName = firstName;
            _LastName = lastName;
            _Company = company;
            _Email = email;
            _Status = status;
            _Result = result;

            _FindCommand = new DelegateCommand(_ => dbMode.Start());

            // The disposal we keep in mind.
            _ToDispose = new[] { 
                passportNr, firstName, lastName, company, email, 
                db, controlsToUser, userToMode, resultToStatus, 
                dbModeToDb, dbToResult, userToResult, resultToPassportNr, 
                resultToFirstName, resultToLastName, 
                resultToCompany, resultToEmail };
        }

        //=//

        private readonly IValidationInputProperty<string> _PassportNr;
        public IValidationInputProperty<string> PassportNr { get { return _PassportNr; } }

        private readonly IValidationInputProperty<string> _FirstName;
        public IValidationInputProperty<string> FirstName { get { return _FirstName; } }

        private readonly IValidationInputProperty<string> _LastName;
        public IValidationInputProperty<string> LastName { get { return _LastName; } }

        private readonly IValidationInputProperty<string> _Company;
        public IValidationInputProperty<string> Company { get { return _Company; } }

        private readonly IValidationInputProperty<string> _Email;
        public IValidationInputProperty<string> Email { get { return _Email; } }

        private readonly ICommand _FindCommand;
        public ICommand FindCommand { get { return _FindCommand; } }

        private readonly IProperty<Employee> _Result;
        public IProperty<Employee> Result { get { return _Result; } }

        private readonly IProperty<LoadingStatus> _Status;
        public IProperty<LoadingStatus> Status { get { return _Status; } }


        // Instead of this in real world we would call an injected service
        private Employee GetFakeEmployee(string passportNr)
        {
            Thread.Sleep(1000);
            if (passportNr == "1111")
                return new Employee(2, "1111", "Me", "No Name", "Fake Company", "me@nomail.org");
            else
                return null;
        }

        //=//

        private readonly IList<IDisposable> _ToDispose;

        public void Dispose()
        {
            _ToDispose.ToList().ForEach(o => o.Dispose()); ;
        }
    }
}
